/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
.COPYRIGHT    (c)  1994   Special Astrophysical Observatory RAS
.IDENT        bgd_sher.c
.LANGUAGE     C
.AUTHOR       Kniazev A.Yu. Laboratory "Structure",  Shergin V.S. InfLab
.KEYWORDS     One-dimensional spectra, backgroung, continuum
.PURPOSE      Automaticly create continuum level for one-dimensional spectra

.COMMENT      Input : IN_A  - input spectra name
		      IN_B  - type of background(a,e,b) and output(r,b)
		      INPUTI(1) - iterations number
		      INPUTI(2) - smoothing mode (if >0 - fitting polynom power)
		      INPUTR(1) - smoothing window size
		      INPUTR(2) - noise level
		      INPUTR(3) - fitting step (if INPUTI(2) > 0)

	      Output: OUT_A - output spectra name
		      OUT_B - pixels weights name

.VERSION      1.0  940222:   Creation AKN and VSher
	      1.1  950711:   VSher - add polynomial smoothing
---------------------------------------------------------------------*/
#include   <stdio.h>
#include   <math.h>
#include   <midas_def.h>

/* float sigma_3p (float *, int);
float sigma_wt (float *, float *, int);
float sigma_amd (float *, int, int); */
float sigma_3p();
float sigma_wt();
float sigma_amd();

main()
{
    static char infile[64], outfile[64], cbuf[16];
    static char wtfile[64] = "weights";
    float *in_data, *out_data, *wt_data;
    int stat, nval, naxis, unit, null, imnoa, imnob, imnow;
    char ident[TEXT_LEN+1], cunit[16*2+1], text[TEXT_LEN];
    int length, i, n;
    static char sp_type[2];
    float port, dt, noise, fstep, noise3, noisew;
    int inputi[3], iter_cnt, smode;

    int npix[2];
    double  start[2], step[2];

    SCSPRO("bgd_sher");

	/* Input parameters */
    stat = SCKGETC("IN_A",1,60,&nval,infile);               /* get input frame   */
    stat = SCKGETC("IN_B",1L,2L,&nval, sp_type);              /*   */
    stat = SCKRDI("INPUTI",1L,3L,&nval,inputi,&unit,&null);   /*   */
    stat = SCKRDR("INPUTR",1L,1L,&nval,&port,&unit,&null);    /*   */
    stat = SCKRDR("INPUTR",2L,1L,&nval,&noise,&unit,&null);   /*   */
    stat = SCKRDR("INPUTR",3L,1L,&nval,&fstep,&unit,&null);   /*   */
    iter_cnt = inputi[0];
    smode    = inputi[1];
	/* Output parameters */
    stat = SCKGETC("OUT_A",1,60,&nval,outfile);   /* get output frame name */
    stat = SCKGETC("OUT_B",1,60,&nval,wtfile);    /* get weights frame name */

	/* map input data frame and get some information from descriptors */
    stat = SCIGET(infile, D_R4_FORMAT, F_I_MODE, F_IMA_TYPE, 2L,
	   &naxis, npix, start, step, ident, cunit, (char **)&in_data, &imnoa);
    if (naxis > 2) SCETER(3L,"Too many dimensions...");

	/* open wrk-frame and write standard descriptors */
    stat = SCIPUT(wtfile, D_R4_FORMAT, F_O_MODE, F_IMA_TYPE,
	   naxis, npix, start, step, ident, cunit, (char **)&wt_data, &imnow);

	/* open output frame and write standard descriptors */
    stat = SCIPUT(outfile, D_R4_FORMAT, F_O_MODE, F_IMA_TYPE,
	   naxis, npix, start, step, ident, cunit, (char **)&out_data, &imnob);

    if (naxis == 1) npix[1] = 1;
    length   = npix[0];
    dt       = (float) step[0];
    if (fstep < 1.e-30)  fstep = port/7.;
    else if (fstep > port/2.) fstep = port/2.;
    i = fstep/dt;
    fstep = (i<1)? dt : i*dt;
	/* Now we call the main subroutine */
    SCTPUT("Background fitting program:");
    SCTPUT("===========================");
    sprintf(text,"Input image:          %s",infile);
    SCTPUT(text);
    sprintf(text,"Output image:         %s",outfile);
    SCTPUT(text);
    sprintf(text,"Smoothing window:     %f",port);
    SCTPUT(text);
    sprintf(text,"Noise estimation:     %f",noise);
    SCTPUT(text);
    sprintf(text,"Number of iterations: %d",iter_cnt);
    SCTPUT(text);
    switch (sp_type[0]) {
    case 'e':           /* emission */
	    SCTPUT("Type of continuum: baseline for emission type spectra");
	    break;
    case 'a':           /* absorbtion */
	    SCTPUT("Type of continuum: baseline for absorption type spectra");
	    break;
    case 'b':           /* both */
    default :
	    SCTPUT("Type of continuum: baseline");
	    break;
    }
    if( smode < 0 )
	    SCTPUT("Method: sliding window average");
    else if( smode == 0 )
	    SCTPUT("Method: convolution with gauss-like window");
    else if( smode == 1 ) {
	    sprintf(text,"Method: linear fitting with step %f", fstep);
	    SCTPUT(text);
    } else if( smode == 2 ) {
	    sprintf(text,"Method: parabolic fitting with step %f", fstep);
	    SCTPUT(text);
    } else {
	    sprintf(text,"Method: polynomial fitting of power %d with step %f",smode,fstep);
	    SCTPUT(text);
    }
    SCTPUT("===========================");

	/* For two-dimensional array of one-dimensional spectra */
    SCTPUT("***************************************");
    for (i=0;i<npix[1];i++) {
	sprintf(text,"One-dimensional spectrum number: %d",i+1);
	SCTPUT(text);
	sprintf(text,"3-points sigma estimation: %f",
	       noise3=sigma_3p(&in_data[npix[0]*i], length));
	SCTPUT(text);
	n = (int)(port/dt + 0.5);
	sprintf(text,"AMD sigma estimation: %f",
	       sigma_amd(&in_data[npix[0]*i], n, length/n));
	SCTPUT(text);
/**/
	stat = bgd_calc (&in_data[npix[0]*i], &out_data[npix[0]*i],
			 &wt_data[npix[0]*i], length, port, dt, sp_type,
			 noise, iter_cnt, smode, fstep);
	if (stat == 1) SCTPUT("bgd_sher: Not enough memory for working buffer");
	sprintf(text,"Weighted sigma estimation: %f",
	       noisew=sigma_wt(&in_data[npix[0]*i], &out_data[npix[0]*i],
			       &wt_data[npix[0]*i], length, sp_type[1]));
	SCTPUT(text);
    }
close: /* Close all */
    stat = SCDCOP(imnoa,imnob,1L,cbuf);     /*  copy all descriptors */
    SCFCLO(imnoa);
    SCFCLO(imnob);
    SCFCLO(imnow);
    SCSEPI();
}

/*-------------------------------------------------------------------*/
/* bgd_calc() - some strange algorithm with long-old-difficult fate: */
/* created for spectrum continuum detection by VSher at ~1982?       */
/*                                   (FORTRAN-4, RT-11, pdp11/4)     */
/* applied for radio-records background fitting by VVV & Co at ~1983?*/
/*                                   (FORTRAN-4, RT-11/TSX, pdp11/4) */
/* rewriten (use files instead of memory buffers) by MLM at ~1987    */
/*                                   (C, Unix/BSD2.9, pdp11/24)      */
/* port to PC by MLM, ~1990          (MS-C, MSDOS, AT286)            */
/* changed by VO, 1/6/93, return to work with memory buffers         */
/*                                   (C, SCO Xenix, AT286/386)       */
/* changed by VO, 18/2/94: to use as subroutine (Interactive Unix, AT386) */
/* AKN & VSher, 23/2/94: fix VO-bugs and prepare for use with MIDAS  */
/* VSher,       11/7/95:  add polynomial smoothing and scaling noise */
/*-------------------------------------------------------------------*/
#define FNULL  ((float *)0)
#define FMAX   (1.e+30)
#define FZERO  (1.e-20)

/* common block: */
static int io_len;            /* in/out data length */
static float rsb, rse;        /* first/last (average) values in record */
static int iport;             /* pixels in window */
static float *f_src;          /* source record */
static float *w_data;         /* current weights of each pixel */
static float *f_fil;          /* current background estimation */


int bgd_calc(in_data, out_data, wt_data, length, port, dt, type, noise,
		     iter_cnt, smode, fstep)
float *in_data;
float *out_data;
float *wt_data;
int length;         /* array's length              */
float port;         /* window of smoothing         */
float dt;           /* step in record (CDELT1)     */
char type[];        /* type[0] 'a'-absorption, 'e'-emission, 'b'-both */
		    /* type[1] 'b'-bgd out; 'r'-subtraction (residuals) out*/
float noise;        /* noise level                 */
int iter_cnt;       /* iterations number           */
int smode;          /* if smode<0 - quick-rough bgd; (sliding average) */
		    /*         =0 - slow bgd; convolution with gauss-like window */
		    /*         >0 - polynomial bgd; smode=power of polynoms*/
float fstep;        /* step of polynoms calculation ( if smode > 0) << port! */
{
	register int i;
	float ane, ana, val;
	int pstep;

/* common block: */
	f_src  = in_data;    /* source record */
	w_data = wt_data;    /* current weights of each pixel */
	f_fil  = out_data;   /* current background estimation */
	io_len = length;     /* in/out data length */
	iport  = port / dt;  /* pixels in window */
			     /* first/last (average) values in record: */

	rsb = ((float)f_src[0]+(float)f_src[1]+(float)f_src[2])/3.0;
	rse = ((float)f_src[length-1] +
	       (float)f_src[length-2] + (float)f_src[length-3])/3.0;
	pstep = fstep / dt;

	switch (type[0]) {
	case 'e':           /* emission */
		ane = noise;
		ana = FMAX;
		break;
	case 'a':           /* absorption */
		ane = FMAX;
		ana = noise;
		break;
	case 'b':           /* both */
	default :
		ane = noise;
		ana = noise;
		break;
	}

	for (i = 0; i < length; i++) {
	       w_data[i] = 1.0;
	}

	while (iter_cnt--) {
		if( smode < 0 )
		    qbgd_fil(f_src, f_fil, w_data, length, iport, rsb, rse);
		else if( smode == 0 )
		    bgd_fil(f_src, f_fil, w_data, length, iport, rsb, rse);
		else if( smode == 1 ) {
		    if(lbgd_fil(f_src, f_fil, w_data, length, iport, rsb, rse, pstep))
		       return(1);
		} else if( smode == 2 ) {
		    if(p2_bgd_fil(f_src, f_fil, w_data, length, iport, rsb, rse, pstep))
		       return(1);
		} else /* smode = power >=3 */ {
		    if(pn_bgd_fil(f_src, f_fil, w_data, length, iport, rsb, rse, pstep, smode))
		       return(1);
		}
		bgd_cut(f_src, f_fil, w_data, length, ane, ana);
	}

	for (i = 0; i < length; i++) {
		val = f_fil[i];
		if (type[1] == 'r')
			 val = in_data[i] - val;
		out_data[i] = val;
	}

	return 0;
}


#define FM  1
#define BM -1
#define WL_SIZE 21

static float wl[WL_SIZE] = {        /* gauss-like smoothing window */
	1.,  .995, .99, .98, .97, .95, .92, .88, .82, .75,
	.65,.54,  .42, .31, .22, .14, .08, .04, .02, .01, 0.01
};
/* static fun1(int, int, float, float *, float *);
static fun2(int, float, float *, float *, float *); */
static fun1();
static fun2();

bgd_fil(f_s, f_f, w_d, length, ip, rb, re)  /* weighted smoothing */
float *f_s, *f_f;
float *w_d;
int length, ip;
float rb, re;
{
	register int i=0;
	register float *fptr;
	float fwstep;

/* common block:   */
/*      f_src = f_s;             source record */
/*      w_data = w_d;            current weights of each pixel */
/*      f_fil = f_f;             current background estimation */
/*      io_len = length;         in/out data length */
/*      rsb = rb; rse = re;      first/last (average) values in record */
/*      iport = ip;              pixels in window */

	fwstep = iport / 40.;
	fptr = f_fil;
	for (i = 0; i < length; ++i ) {
		float f0 = 0.0, s0 = 0.0;

		fun1(i + 1, FM, fwstep, &f0, &s0);
		fun1(i, BM, fwstep, &f0, &s0);

		*fptr++ = f0/s0;
	}
}

static fun1(i, ist, wstep, f0, s0)   /* half of convolution sums */
register int i;               /* center location  */
int ist;                      /* step (direction) */
float wstep, *f0, *s0;
{
	float f1 = 0.0, s1 = 0.0, s = 0.0;
	register int l = 0;

	while (l < WL_SIZE-1) {
		float w0, w1;

		w0 = wl[l] * wstep;
		w1 = wl[l];
		s += w0;
		while (s1 < s ) {
			fun2(i, w1, &s, &s1, &f1);
			i += ist;
		}
		l++;
	}
	*s0 += s1;
	*f0 += f1;
}

static fun2(i, w, s, s1, f1)    /* next weighted point to sum */
register int i;
float w, *s, *s1, *f1;
{
	register int k;
	float rsj, val;

	if (i >= 0 && i < io_len) {
		k = i;
		rsj = f_src[k];
	}
	else  if ( i < 0  ) {        /* "double-mirror" edges maintence */
		k = -i;
		rsj = rsb + rsb - f_src[k];
	}
	else if ( i >= io_len) {
		k = io_len + io_len - i - 1;
		rsj = rse + rse - f_src[k];
	}

	val = w_data[k];
	if ( val > 0.00001) {
		float q;
		q = w * val;
		*s1 += q;
		*f1 += rsj * q;
	}
	*s -= w * 0.3 * (1. - val);    /* some magic....*/
}

qbgd_fil(f_s, f_f, w_d, length, ip, rb, re) /* sliding window smoothing */
float *f_s, *f_f;
float *w_d;
int length, ip;
float rb, re;
{
	register int i, k;
	int ib, ie, ie0;
	register float *fptr;
	float s1, s2, w, rsj;

/* common block:   */
/*      f_src = f_s;             source record */
/*      w_data = w_d;            current weights of each pixel */
/*      f_fil = f_f;             current background estimation */
/*      io_len = length;         in/out data length */
/*      rsb = rb; rse = re;      first/last (average) values in record */
/*      iport = ip;              pixels in window */

	fptr = f_fil;
	/* set starting position of sliding window:                      */
			     /*    / window=iport  \                     */
	i  = -iport;         /*  /---------+---------\----------|---.... */
	ib = i+i/2;          /* ib         i         ie         |        */
	ie = ie0 = i/2;      /*         -iport     -iport/2     0        */

	for(s1=s2=0.0; i<length; i++,ib++,ie++) {
	    if (ie >= 0 && ie < length) {
		    k = ie;
		    rsj = f_src[k];
	    }
	    else  if ( ie < 0  ) {      /* "double-mirror" edges maintence */
		    k = -ie;
		    rsj = rsb + rsb - f_src[k];
	    }
	    else /* if ( ie >= length) */ {
		    k = length + length - ie - 1;
		    rsj = rse + rse - f_src[k];
	    }
	    s2 += (w = (w_data[k]>0.001)? w_data[k] : 0.001);
	    s1 +=  rsj * w;
	    if(ib >= ie0) {
		if (ib >= 0) {
			k = ib;
			rsj = f_src[k];
		}
		else /* if (ib < 0) */ { /* "double-mirror" edges maintence */
			k = -ib;
			rsj = rsb + rsb - f_src[k];
		}
		s2 -= (w = (w_data[k]>0.001)? w_data[k] : 0.001);
		s1 -=  rsj * w;
	    }
	    if(i>=0) {
		*fptr++ = s1/s2;
	    }
	}
}

static float *ww=NULL;
static int lw=0;

static int mk_wnd(ip)            /* make gauss-window (weights) */
int ip;                          /* pixels in window            */
{
   register int i;
   double f;

   if(ip > lw) {
       if(ww) free(ww);
       if((ww = (float *)calloc(ip, sizeof(*ww))) == NULL)  return (-1);
   }
   if(ip != lw) {
       lw = ip;
       for(i=0; i < ip; i++ ) {
	   f = (double)i/((double)ip/3.0);
	   ww[i] = exp(-f*f);
       }
   }
   return(0);
}

lbgd_fil(f_s, f_f, w_d, length, ip, rb, re, lstep) /* linear weighted smoothing */
float *f_s;                      /* source record                 */
float *f_f;                      /* current background estimation */
float *w_d;                      /* current weights of each pixel */
int length;                      /* in/out data length  */
int ip;                          /* pixels in window    */
float rb, re;                    /* first/last (average) values in record */
int lstep;                       /* step of LSQ-lines calculation     */
{
   register int j,k,l;
   int i0,i1, i, ib,ie;
   double sx0, sx1, sx2, sfx0, sfx1;
   double x0,x1,x, w, f, d, da0, db0, da1, db1;

   if(ip > lw)
       if( mk_wnd(ip) )  return (-1);

   if (lstep > ip/2) lstep = ip/2;
   if (lstep < 1)    lstep = 1;

   da0=db0=da1=db1=0.0;
   x0 = x1 = i0 = i1 = -lstep;
   for(i = 0; i < length; ++i ) {
       if(i > i1) {
	   da0=da1; db0=db1;
	   i0=i1;  x0=x1;
	   x1 = (i1 += lstep);
	   sx0=sx1=sx2=sfx0=sfx1=0.0;
	   ib = i1-ip+1;
	   ie = i1+ip-1;
	   for(j=ib; j < ie; j++ ) {
	       l = (j>i1)? j-i1 : i1-j;
	       if (j >= 0 && j < length) {
		       k = j;
		       f = f_s[k];
	       }
	       else  if ( j < 0  ) {      /* "double-mirror" edges maintence */
		       k = -j;
		       f = rb + rb - f_s[k];
	       }
	       else /* if ( j >= length) */ {
		       k = length + length - j - 1;
		       f = re + re - f_s[k];
	       }
	       w = ww[l] * w_d[k];
	       x = j;
	       sx0 += w;
	       sfx0 += w * f;
	       sx1 += x * w;
	       sfx1 += w * x * f;
	       sx2 += x*x * w;
	   }
	   d   = sx0*sx2 - sx1*sx1;
	   if(d<FZERO) d=FZERO;
	   da1 = (sx0*sfx1 - sx1*sfx0) / d;
	   db1 = (sfx0*sx2 - sfx1*sx1) / d;
       }
       x = i;
       if (i==i0)
	   f_f[i] = da0*x + db0;   /* will never occur?*/
       else if (i==i1)
	   f_f[i] = da1*x + db1;
       else /* i0 < i < i1 */ {
	   f = (x1-x)/(x1-x0);
	   f_f[i] = (da0*x + db0) * f +
		    (da1*x + db1) * (1.0 - f);
       }
   }
   return(0);
}

p2_bgd_fil(f_s, f_f, w_d, length, ip, rb, re, pstep) /* parabolic weighted smoothing */
float *f_s;                      /* source record                 */
float *f_f;                      /* current background estimation */
float *w_d;                      /* current weights of each pixel */
int length;                      /* in/out data length  */
int ip;                          /* pixels in window    */
float rb, re;                    /* first/last (average) values in record */
int pstep;                       /* step of LSQ-parabolas calculation     */
{
   register int j,k,l;
   int i0,i1, i, ib,ie;
   double sx0, sx1, sx2, sx3, sx4, sfx0, sfx1, sfx2;
   double x0,x1,x, w, f, d, da0, db0, dc0, da1, db1, dc1;

   if(ip > lw)
       if( mk_wnd(ip) )  return (-1);

   if (pstep > ip/2) pstep = ip/2;
   if (pstep < 1)    pstep = 1;

   da0=db0=dc0=da1=db1=dc1=0.0;
   x0 = x1 = i0 = i1 = -pstep;
   for(i = 0; i < length; ++i ) {
       if(i > i1) {
	   da0=da1; db0=db1; dc0=dc1;
	   i0=i1;  x0=x1;
	   x1 = (i1 += pstep);
	   sx0=sx1=sx2=sx3=sx4=sfx0=sfx1=sfx2=0.0;
	   ib = i1-ip+1;
	   ie = i1+ip-1;
	   for(j=ib; j < ie; j++ ) {
	       l = (j>i1)? j-i1 : i1-j;
	       if (j >= 0 && j < length) {
		       k = j;
		       f = f_s[k];
	       }
	       else  if ( j < 0  ) {      /* "double-mirror" edges maintence */
		       k = -j;
		       f = rb + rb - f_s[k];
	       }
	       else /* if ( j >= length) */ {
		       k = length + length - j - 1;
		       f = re + re - f_s[k];
	       }
	       w = ww[l] * w_d[k];
	       x = j;
	       sx0 += w;
	       sfx0 += w * f;
	       sx1 += x * w;
	       sfx1 += w * x * f;
	       x *= j;
	       sx2 += x * w;
	       sfx2 += w * x * f;
	       x *= j;
	       sx3 += x * w;
	       x *= j;
	       sx4 += x * w;
	   }
	   d = sx0*sx2*sx4+2.0*sx1*sx2*sx3-sx2*sx2*sx2-sx0*sx3*sx3-sx1*sx1*sx4;
	   da1 = sx0*sx2*sfx2 + sx1*sfx0*sx3 + sx1*sx2*sfx1 -
		 sx2*sx2*sfx0 - sx0*sfx1*sx3 - sx1*sx1*sfx2;
	   db1 = sx0*sfx1*sx4 + sx1*sx2*sfx2 + sfx0*sx2*sx3 -
		 sx2*sfx1*sx2 - sx0*sx3*sfx2 - sfx0*sx1*sx4;
	   dc1 = sfx0*sx2*sx4 + sx1*sfx2*sx3 + sfx1*sx2*sx3 -
		 sfx2*sx2*sx2 - sfx0*sx3*sx3 - sx1*sfx1*sx4;
	   if(d<FZERO) d=FZERO;
	   da1 /= d;
	   db1 /= d;
	   dc1 /= d;
       }
       x = i;
       if (i==i0)
	   f_f[i] = (da0*x + db0)*x + dc0;   /* will never occur?*/
       else if (i==i1)
	   f_f[i] = (da1*x + db1)*x + dc1;
       else /* i0 < i < i1 */ {
	   f = (x1-x)/(x1-x0);
	   f_f[i] = ((da0*x + db0)*x + dc0) * f +
		    ((da1*x + db1)*x + dc1) * (1.0 - f);
       }
   }
   return(0);
}

#define N_MAX 15                 /* maximum power of fitting polynom */

static double pn_comp(p, x, n)   /* polynom evaluation   */
double p[N_MAX+1],               /* n+1 polynom coefficients */
       x;                        /* argument */
int    n;                        /* polynom's power */
{
   register int j, m=n+1;
   double f = p[0];
   for(j = 1; j < m; j++ )
       f = f * x + p[j];
   return(f);
}

pn_bgd_fil(f_s, f_f, w_d, length, ip, rb, re, pstep, n) /* polynomial weighted smoothing */
float *f_s;                      /* source record                 */
float *f_f;                      /* current background estimation */
float *w_d;                      /* current weights of each pixel */
int length;                      /* in/out data length  */
int ip;                          /* pixels in window    */
float rb, re;                    /* first/last (average) values in record */
int pstep;                       /* step of LSQ-polynoms calculation     */
int n;                           /* fitting polynom power */
{
   register int j,k,l;
   int i0,i1, i, ib,ie, m, ns, ms;
   static double a[N_MAX+1][N_MAX+1],  b[N_MAX+1];
   static double sx[2*N_MAX+1],  sfx[N_MAX+1];
   static double d0[N_MAX+1],    d1[N_MAX+1];
   double x1, x, w, f, xs, xstep;

   static int solve_equat(double [][], double *, double *, int);

   if(ip > lw)
       if( mk_wnd(ip) )  return (-1);

   if (pstep > ip/2) pstep = ip/2;
   if (pstep < 1)    pstep = 1;
   if (n > N_MAX) n = N_MAX;
   if (n > ip/3)  n = ip/3;
   if (n < 1 )    n = 1;
   m = n+1;
   ns = n*2;
   ms = ns+1;
   xs = 100./(double)length;
   xstep = pstep * xs;

   for(j = 0; j < m; j++ )  d0[j] = d1[j] =0.0;

   i0 = i1 = -pstep;
   for(i = 0; i < length; i++ ) {
       if(i > i1) {
	   for(j = 0; j < m; j++ )  d0[j] = d1[j];
	   i0=i1;
	   i1 += pstep;
	   for(j = 0; j < ms; j++ ) sx[j]  = 0.0;
	   for(j = 0; j < m; j++ )  sfx[j] = 0.0;
	   ib = i1-ip+1;
	   ie = i1+ip-1;
	   for(j=ib; j < ie; j++ ) {
	       l = (j>i1)? j-i1 : i1-j;
	       if (j >= 0 && j < length) {
		       k = j;
		       f = f_s[k];
	       }
	       else  if ( j < 0  ) {      /* "double-mirror" edges maintence */
		       k = -j;
		       f = rb + rb - f_s[k];
	       }
	       else /* if ( j >= length) */ {
		       k = length + length - j - 1;
		       f = re + re - f_s[k];
	       }
	       w = ww[l] * w_d[k];
	       x1 = x = (j-i0) * xs;
	       sx[0]  += w;
	       sfx[0] += w * f;
	       for(l = 1; l < m; l++ ) {
		   sx[l]  += x * w;
		   sfx[l] += w * x * f;
		   x *= x1;
	       }
	       for(l = m; l < ms; l++ ) {
		   sx[l] += x * w;
		   x *= x1;
	       }
	   }
	   for(j = 0; j < m; j++ ) {
	       for(l = 0; l < m; l++ ) {
		   a[j][l] = sx[n*2-j-l];
	       }
	       b[j] = sfx[m-j-1];
	   }
	   if (solve_equat(a, b, d1, m)) {
	       printf("\nLinear equations solving problem: wrong matrix...\n");
	       for(j = 0; j < m; j++ )  d1[j] = d0[j];
	   }
       }
       x = (i-i0) * xs;
       if (i==i0)
	   f_f[i] = pn_comp(d0,x+xstep,n);   /* will never occur?*/
       else if (i==i1)
	   f_f[i] = pn_comp(d1,x,n);
       else /* i0 < i < i1 */ {
	   f = (xstep-x)/(xstep);
	   f_f[i] = pn_comp(d0,x+xstep,n) * f + pn_comp(d1,x,n) * (1.0 - f);
       }
   }
   return(0);
}

static float wn[10] = {        /* cutting weigth */
	1., .99, .97, .89, .7, .35, .11, .04, .01, .0 };
static float f_avr = 0.0;

bgd_cut(f_src, f_fil, w_data, length, ane, ana)
float *f_src, *f_fil;
float *w_data;
int length;
float ane, ana;
{
	register int i;
	register int l;
	float s, val, scale;

	if(ana<FZERO) ana=FZERO;
	if(ane<FZERO) ane=FZERO;

	if(f_avr == 0.0) {
	    for ( s=0.,i = 0; i < length; i++)
		s += f_fil[i];
	    f_avr = s/length;
	}
	for ( i = 0; i < length; i++) {
		val = f_src[i] - f_fil[i];
		scale = f_fil[i] / f_avr;
		scale = (scale > 0.0)? sqrt((double)scale) : 1.0;
		if (fabs(val) < FZERO)
			{l = 0; s = 0.0;}
		else if (val <  0.0)
			l = (int)(s = -5. * (val / (ana*scale)));
		else if ( val > 0.0 )
			l = (int)(s = 5. * (val / (ane*scale)));
		if (l <= 0 && s < 0.5 )
			val = wn[0];
		else if (l >= 0 && l <= 8)
			val = wn[l] + (wn[l+1] - wn[l]) * (s - l);
		else if (l >= 9)
			val = wn[9];
		w_data[i] = val;
	}
}

/* VSher & HBH, various SIGMA estimation methods */

float sigma_3p (in_data, length)
float *in_data;
int length;
{
	register i;
	double qub, s=0.0;

	for (i=1;i<(length-1);i++) {
	  qub = in_data[i] - (in_data[i-1]+in_data[i+1])/2.;
	  s = s + qub*qub;
	}
	return (float) sqrt(s/(length-3));
}

float sigma_wt ( src_data, out_data, weight, length, otype)
float *src_data, *out_data, *weight;
int length;
char otype;
{
	register i;
	double qub, s=0.0, sw=0.0;

	for (i=1;i<(length-1);i++) {
	  qub = (otype=='b')? src_data[i] - out_data[i] : out_data[i];
	  s  = s  + qub*qub*weight[i];
	  sw = sw + weight[i];
	}
	return (float) sqrt(s/(sw-1.0));
}

static int fcmp(f1,f2)     /* compare two floats for qsort() */
float *f1, *f2;
{
	return( ((*f1 - *f2) >= 0.0)? 1 : -1 );
}

/* "Quicker-sort algorithm" from system library       */
static float fqsrt(arr,k)  /*  mdn=fqsrt(arr,k)       */
float *arr;                /*  sort k elements in arr */
register int k;            /*  returns median of arr  */
{
	register int i = k/2;
	int fcmp();
	qsort(arr, k, sizeof(float), fcmp);
	return( (k&01)? arr[i] : (arr[i-1]+arr[i])/2.0);
}

/* HBH favourite method: */
/* AMD (AbsoluteMedianDeviation) - robust estimation of "sigma"*/
float sigma_amd(vect,n,nn)
int n;                          /* n - size of window */
int nn;                         /* nn - number of windows in vector*/
float *vect;                    /* vect -  array of n*nn elements */
{
	float amd, med, b;
	float *buf0, *pb0, *pvect, *buf1, *pb1;
	register int i,j;

	if((buf0 = (float *)calloc(n,sizeof(*pb0))) == NULL)
	    return 0.0;
	if((buf1 = (float *)calloc(nn,sizeof(*pb1))) == NULL) {
	    free((char*)buf0);
	    return 0.0;
	}
	for(pb1=buf1,i=0; i<nn; i++) {
		pvect = vect+i*n;
		for(pb0=buf0,j=0; j<n; j++)
		      *pb0++ = *pvect++;
		med = fqsrt(buf0, n);
		for(pb0=buf0,j=0; j<n; j++) {
		       b = (*pb0 - med);
		       *pb0++ = (b > 0.0) ? b : -b;
		}
		*pb1++ = fqsrt(buf0,n);
	}
	amd = fqsrt(buf1,nn)/(0.674*sqrt((double)(n-1.)/n));
	free((char*)buf1);
	free((char*)buf0);
	return(amd);
}

/*  Oleg V.Verkhodanov <vo@sao.stavropol.su>
 *  compact method of equation sets decision
 */
#define ABS(x)  ((x) < 0 ? -(x): (x))
#define SMALL   1.e-100

static int solve_equat(a,b,x,n)
double a[N_MAX+1][N_MAX+1], /* square matrix of coefficients [n]x[n] */
       *b,      /*    array of equations meanings(free parameters)  */
       *x;      /*    matrix of variables                           */
int    n;       /*    dimension of arrays                           */
{
	register i,j;
	int k, ret=0;
	double   d;

	for(k=0; k < n-1; k++) {
		/* make sure  a_kk != 0 */
		for(i=k; i < n; i++)
		      if(ABS(a[i][k]) > SMALL)
			    break;
		if ( i == n ) {
		      /*printf("\nLinear equations solving problem: wrong matrix...\n");*/
			ret = -1;
			/* Try to prevent  FP error */
			for(j=k; j < n; j++)
				a[j][j] += SMALL*10.;
		} else
			if ( ABS(a[k][k]) < SMALL) {
				for(j=k; j < n; j++)
					a[k][j] += a[i][j];
				b[k] += b[i];
			}

		/* make a_ik = 1 */
		for(i=k; i<n; i++) {
			d = a[i][k];
			if(ABS(d) > SMALL) {
				for(j=k; j<n; j++)
				       a[i][j] /= d;
				b[i] /= d;
			}
		}
		/* make matrix triangular */
		for(i=k+1; i<n; i++) {
			if (ABS(a[i][k]) > SMALL ) {
				for(j=k; j<n; j++)
				       a[i][j] -= a[k][j] ;
				b[i] -= b[k];
			}
		}
	}

	x[n-1] = b[n-1] / a[n-1][n-1];
	for(i=n-2; i>=0; i--) {
		d = 0.;
		for(j=i+1; j<n; j++)
			d += x[j] * a[i][j];
		x[i] = b[i] - d;
	}
	return (ret);
}


